use std::fs::{read_link, DirEntry};
use std::path::PathBuf;

use eframe::egui::{Id, Key};
use eframe::{egui, Frame, NativeOptions};
use egui_extras::{Column, TableBuilder};

fn main() {
	eframe::run_native(
		"File Exploder",
		NativeOptions::default(),
		Box::new(|cc| Box::new(FileGui::new(cc))),
	)
	.unwrap();
}

struct FileGui {
	home: String,
	current_dir: PathBuf,
	loaded_dir: PathBuf,
	entries: Vec<Entry>,
	path_input: String,
	path_input_id: Id,
	search_input: String,
	search_input_id: Id,
}

struct Entry {
	name: String,
	path: PathBuf,
	filetype: FileType,
	visible: bool,
}

#[derive(PartialEq)]
enum FileType {
	Dir,
	File,
	SymLinkFile,
	SymLinkDir,
	SymLinkMissing,
	Unknown, // e.g. device files on linux
}

impl FileGui {
	pub fn new(_cc: &eframe::CreationContext<'_>) -> Self {
		let home = std::env::var("HOME").unwrap();
		let current_dir = std::env::current_dir().unwrap();
		let mut s = Self {
			home,
			current_dir,
			loaded_dir: PathBuf::new(),
			entries: Vec::new(),
			path_input: String::new(),
			path_input_id: Id::null(),
			search_input: String::new(),
			search_input_id: Id::null(),
		};
		s.override_input();
		s.refresh_entries();
		s
	}

	fn override_input(&mut self) {
		self.path_input = self
			.current_dir
			.to_str()
			.unwrap()
			.replace('\\', "/")
			.replace(&self.home, "~");
	}

	fn apply_input(&mut self) {
		let path = self.path_input.replace('~', &self.home).replace('\\', "/");
		self.current_dir = PathBuf::from(path);
	}

	fn refresh_entries(&mut self) {
		if !self.current_dir.is_dir() {
			self.current_dir = self.loaded_dir.clone();
		}
		if !self.current_dir.is_dir() {
			return;
		}
		self.entries = self
			.current_dir
			.read_dir()
			.unwrap()
			.filter_map(|dir_entry| dir_entry.ok().map(Entry::new))
			.flatten()
			.collect();
		self.entries
			.sort_by(|a, b| b.is_dir().cmp(&a.is_dir()).then(a.name.cmp(&b.name)));
		self.refresh_filter();
		self.loaded_dir = self.current_dir.clone();
	}

	fn refresh_filter(&mut self) {
		for entry in &mut self.entries {
			entry.visible = entry.name.contains(&self.search_input);
		}
	}

	fn to_parent(&mut self) {
		if let Some(path) = self.current_dir.parent() {
			self.current_dir = path.to_owned();
			self.refresh_entries();
			self.override_input();
		}
	}

	fn open(&mut self, index: usize) {
		let entry = &self.entries[index];
		match entry.filetype {
			FileType::Dir | FileType::SymLinkDir => {
				self.current_dir = entry.path.clone();
				self.refresh_entries();
				self.override_input();
			}
			_ => {
				opener::open(&entry.path).unwrap();
			}
		}
	}

	fn key_inputs(&mut self, ctx: &egui::Context) {
		if ctx.input(|i| i.key_pressed(Key::F) && i.modifiers.ctrl) {
			ctx.memory_mut(|m| m.request_focus(self.search_input_id));
		}
		if ctx.input(|i| i.key_pressed(Key::L) && i.modifiers.ctrl) {
			ctx.memory_mut(|m| m.request_focus(self.path_input_id));
		}
		if ctx.input(|i| i.key_pressed(Key::R) && i.modifiers.ctrl) {
			self.refresh_entries();
		}
		if ctx.input(|i| i.key_pressed(Key::ArrowUp) && i.modifiers.alt) {
			self.to_parent();
		}
	}
}

impl eframe::App for FileGui {
	fn update(&mut self, ctx: &egui::Context, frame: &mut Frame) {
		self.key_inputs(ctx);
		egui::TopBottomPanel::top("top_bar").show(ctx, |ui| {
			ui.horizontal(|ui| {
				if ui.button("⏶").clicked() {
					self.to_parent();
				}
				if ui.button("⟳").clicked() {
					self.refresh_entries();
				}
				let path_bar = ui.text_edit_singleline(&mut self.path_input);
				self.path_input_id = path_bar.id;
				if path_bar.lost_focus() {
					if ui.input(|i| i.key_pressed(egui::Key::Enter)) {
						self.apply_input();
						self.refresh_entries();
					} else if ui.input(|i| i.key_pressed(egui::Key::Escape)) {
						self.override_input();
					}
				}
				let filter_input = ui.text_edit_singleline(&mut self.search_input);
				self.search_input_id = filter_input.id;
				if filter_input.changed() {
					self.refresh_filter();
				}

				if let Some(filename) = self.loaded_dir.file_name() {
					ui.label(filename.to_str().unwrap());
				}
			});
		});
		egui::CentralPanel::default().show(ctx, |ui| {
			TableBuilder::new(ui)
				.striped(true)
				.max_scroll_height(frame.info().window_info.size.y)
				.column(Column::auto())
				.column(Column::remainder())
				.body(|mut body| {
					let mut to_open = None;
					for (row_index, entry) in self.entries.iter().enumerate() {
						if entry.visible {
							body.row(18.0, |mut row| {
								row.col(|ui| {
									ui.label(entry.filetype.icon());
								});
								row.col(|ui| {
									let f = ui.button(&entry.name);
									if f.clicked() {
										to_open = Some(row_index);
									}
									// if f.has_focus() && ui.input(|i| i.key_pressed(Key::Delete)) {}
								});
							});
						}
					}
					if let Some(row_index) = to_open {
						self.open(row_index);
					}
				});
		});
	}
}

impl Entry {
	fn new(dir_entry: DirEntry) -> Option<Self> {
		let name = dir_entry.file_name().into_string().ok()?;
		let path = dir_entry.path();
		let filetype = dir_entry.file_type().ok()?;
		let filetype = if filetype.is_dir() {
			FileType::Dir
		} else if filetype.is_file() {
			FileType::File
		} else if filetype.is_symlink() {
			let p = read_link(&path).ok()?;
			if p.is_dir() {
				FileType::SymLinkDir
			} else if p.exists() {
				FileType::SymLinkFile
			} else {
				FileType::SymLinkMissing
			}
		} else {
			FileType::Unknown
		};

		Some(Self {
			visible: true,
			name,
			path,
			filetype,
		})
	}

	fn is_dir(&self) -> bool {
		self.filetype == FileType::Dir || self.filetype == FileType::SymLinkDir
	}
}

impl FileType {
	fn icon(&self) -> &str {
		match self {
			FileType::Dir => "🗁",
			FileType::File => "🗋",
			FileType::SymLinkFile => "⮩",
			FileType::SymLinkDir => "⎆",
			FileType::SymLinkMissing => "⚠",
			FileType::Unknown => "？",
		}
	}
}
